#include <boost/algorithm/clamp.hpp>

#include <SimoxUtility/math/convert/mat4f_to_pos.h>
#include <SimoxUtility/math/convert/mat4f_to_quat.h>
#include <SimoxUtility/math/compare/is_equal.h>

#include <SimoxUtility/json.h>
//#include <RobotAPI/libraries/core/json_conversions.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>

#include <VirtualRobot/MathTools.h>
#include "../utils.h"
#include "FTSensor.h"


namespace armarx::control::common::ft
{

void FTSensor::initialize(
        const VirtualRobot::RobotNodeSetPtr& rns,
        const RobotUnitPtr& robotUnit,
        const FTConfig& c
        )
{
    const SensorValueBase* svlf = robotUnit->getSensorDevice(c.sensorName)->getSensorValue();
    forceSensor = svlf->asA<SensorValueForceTorque>();

    ARMARX_CHECK_NOT_NULL(rns);
    sensorNode = rns->getRobot()->getSensor(c.sensorName)->getRobotNode();

    if (c.sensorName == "FT R")
    {
        tcpMass = calibDefalts.tcpMassRight;
        tcpCoMInFTSensorFrame = calibDefalts.tcpCoMInFTFrameRight;
    }
    else if (c.sensorName == "FT L")
    {
        tcpMass = calibDefalts.tcpMassLeft;
        tcpCoMInFTSensorFrame = calibDefalts.tcpCoMInFTFrameLeft;
    }

    filteredForce.setZero();
    filteredTorque.setZero();
    filteredForceInRoot.setZero();
    filteredTorqueInRoot.setZero();

    tcpGravityCompensation.setZero();
    reset();
}

//FTSensor::FTConfig FTSensor::reconfigure(const std::string &configFileName)
//{
//    nlohmann::json userConfig;
//    std::ifstream ifs{configFileName};
//    ifs >> userConfig;
//    auto c = userConfig["force_torque"];
//    FTConfig config = FTConfig();
//    std::string name       = getValueWithDefault<std::string>(c, "", "sensor_name");
//    if (name != sensorName)
//    {
//        ARMARX_ERROR << "the sensor name: " << name << " is not the same as initialized: " << sensorName << ", please consider to switch controllers instead of reuse";
//    }
//    config.ftFilter        = getValueWithDefault<float>(c, s.ftFilter, "force_filter_coeff");
//    config.deadZoneForce   = getValueWithDefault<float>(c, s.deadZoneForce, "force_dead_zone");
//    config.deadZoneTorque  = getValueWithDefault<float>(c, s.deadZoneTorque, "torque_dead_zone");
//    config.timeLimit       = getValueWithDefault<float>(c, s.timeLimit, "wait_time_for_calibration");
//    s.timeLimit = config.timeLimit;

//    config.enableTCPGravityCompensation = getValueWithDefault<bool>(c, s.enableTCPGravityCompensation, "enable_tcp_gravity_compensation");
//    nlohmann::json compCfg;
//    if (c.find(sensorName) != c.end())
//        auto compCfg = c[sensorName];
//    config.tcpMass = getValueWithDefault<float>(compCfg, s.tcpMass, "tcp_mass");
//    config.tcpCoMInFTSensorFrame = getEigenVecWithDefault(compCfg, s.tcpCoMInFTSensorFrame, "tcp_CoM_in_FT_frame");
//    config.forceBaseline.setZero();
//    config.torqueBaseline.setZero();
//    return config;
//}

void FTSensor::reset()
{
    /// you have to call this method **only** in the first loop of rtRun
    //    s.tcpMass = 0.0f;
    //    s.tcpCoMInFTSensorFrame.setZero();

    forceOffset.setZero();
    torqueOffset.setZero();
    ft.setZero();

    calibrated.store(false);
    timeForCalibration = 0.0f;

    //    Eigen::Matrix4f forceFrameInRoot = sensorNode->getPoseInRootFrame();
    //    forceFrameInTCP.block<3, 3>(0, 0) = currentPose.block<3, 3>(0, 0).transpose() * forceFrameInRoot.block<3, 3>(0, 0);
    //    tcpCoMInTCPFrame = forceFrameInTCP.block<3, 3>(0, 0) * tcpCoMInFTSensorFrame;
}

bool FTSensor::calibrate(const FTConfig& c, double deltaT)
{
    /// this will calibrate the force torque value in FTSensorFrame until timeLimit is reached
    Eigen::Matrix3f sensorOriInRoot = sensorNode->getOrientationInRootFrame();
    forceOffset  = (1 - c.ftFilter) * forceOffset  + c.ftFilter * (forceSensor->force  - sensorOriInRoot.transpose() * tcpGravityCompensation.head(3));
    torqueOffset = (1 - c.ftFilter) * torqueOffset + c.ftFilter * (forceSensor->torque - sensorOriInRoot.transpose() * tcpGravityCompensation.tail(3));

    timeForCalibration = timeForCalibration + deltaT;
    if (timeForCalibration > c.timeLimit)
    {
        calibrated.store(true);
        /// rt is ready only if the FT sensor is calibrated, and the rt2interface buffer is updated (see above)
        ARMARX_IMPORTANT << "FT calibration done, (RT is ready!) \nforceOffset:\n" << forceOffset << "\ntorqueOffset:\n" << torqueOffset << "\ntcpGravityCompensation:\n" << tcpGravityCompensation;
    }
    return calibrated.load();
}


//void FTSensor::enableGravityCompensation(const float mass, const Eigen::Vector3f& tcpCoMInFTFrame)
//{
//    enableTCPGravityCompensation.store(true);
//    tcpMass = mass;
//    tcpCoMInFTSensorFrame = tcpCoMInFTFrame;
//}

//void FTSensor::disableGravityCompensation()
//{
//    enableTCPGravityCompensation.store(false);
//    tcpMass = 0.0f;
//    tcpCoMInFTSensorFrame.setZero();
//    tcpGravityCompensation.setZero();
//}

Eigen::Vector6f& FTSensor::compensateTCPGravity(const FTConfig& c)
{
    //    /// update variables from rt buffer
    //    tcpMass = data.tcpMass;
    //    tcpCoMInTCPFrame = forceFrameInTCP.block<3, 3>(0, 0) * data.tcpCoMInFTSensorFrame;

    /// static compensation in root frame
    if (c.enableTCPGravityCompensation)
    {

        Eigen::Vector3f forceVec = tcpMass * gravity;
        Eigen::Vector3f torqueVec = (sensorNode->getOrientationInRootFrame() * tcpCoMInFTSensorFrame).cross(forceVec);
        tcpGravityCompensation << forceVec, torqueVec;
        //        Eigen::Vector3f localGravity = currentPose.block<3, 3>(0, 0).transpose() * gravity;
        //        Eigen::Vector3f localForceVec = tcpMass * localGravity;
        //        Eigen::Vector3f localTorqueVec = tcpCoMInTCPFrame.cross(localForceVec);

        //        tcpGravityCompensation << localForceVec, localTorqueVec;
    }
    else
    {
        tcpGravityCompensation.setZero();
    }
    return tcpGravityCompensation;
}


Eigen::Vector6f& FTSensor::getFilteredForceTorque(const FTConfig& c)
{
    filteredForce  = (1 - c.ftFilter) * filteredForce  + c.ftFilter * forceSensor->force;
    filteredTorque = (1 - c.ftFilter) * filteredTorque + c.ftFilter * forceSensor->torque;

    Eigen::Matrix3f forceOriInRoot = sensorNode->getOrientationInRootFrame();
    filteredForceInRoot  = forceOriInRoot * (filteredForce  - forceOffset)  - tcpGravityCompensation.head(3) - c.forceBaseline;
    filteredTorqueInRoot = forceOriInRoot * (filteredTorque - torqueOffset) - tcpGravityCompensation.tail(3) - c.torqueBaseline;

    for (size_t i = 0; i < 3; ++i)
    {
        if (fabs(filteredForceInRoot(i)) > c.deadZoneForce)
        {
            filteredForceInRoot(i) -= (filteredForceInRoot(i) / fabs(filteredForceInRoot(i))) * c.deadZoneForce;
        }
        else
        {
            filteredForceInRoot(i) = 0;
        }

        if (fabs(filteredTorqueInRoot(i)) > c.deadZoneTorque)
        {
            filteredTorqueInRoot(i) -= (filteredTorqueInRoot(i) / fabs(filteredTorqueInRoot(i))) * c.deadZoneTorque;
        }
        else
        {
            filteredTorqueInRoot(i) = 0;
        }
    }

    ft << filteredForceInRoot, filteredTorqueInRoot;
    return ft;
}
}
