#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <mplib/core/SystemState.h>

#include "MP.h"
#include "../utils.h"


namespace armarx::control::common::mp
{

//MP::MP(const std::string &name, nlohmann::json &userConfig, nlohmann::json &defaultConfig)
MP::MP(const MPConfig& c)
{
    cfg = c;
    mplib::factories::VMPFactory mpfactory;
    mpfactory.addConfig("kernelSize", c.kernelSize);

    resetVMPType(c.mpTypeString);
    ARMARX_INFO << "Creating mp with name: " << c.name << "\n" << VAROUT(c.className) << "\n" << VAROUT(c.mpTypeString) << "\n" << VAROUT(c.mpMode) << "\n" << VAROUT(c.mpStyle);
    std::shared_ptr<mplib::representation::AbstractMovementPrimitive> mp = mpfactory.createMP(vmpType);
    vmp = std::dynamic_pointer_cast<mplib::representation::vmp::PrincipalComponentVMP>(mp);

    if (cfg.viaPoints.size() > 0)
    {
        for (const auto& vp : cfg.viaPoints)
        {
            ARMARX_IMPORTANT << VAROUT(vp.canonicalValue) << " " << VAROUT(common::dVecToString(vp.viaPointValue)) << " " << VAROUT(vmp->getUmin());
            ARMARX_CHECK_GREATER_EQUAL(vp.canonicalValue, 0);
            ARMARX_CHECK_LESS_EQUAL(vp.canonicalValue, 1);
            userDefinedViaPoints.push_back({vp.canonicalValue, vp.viaPointValue});
        }
    }
}

std::string MP::getMPName()
{
    return cfg.name;
}

std::string MP::getClassName()
{
    return cfg.className;
}

void MP::start()
{
    start(std::vector<double>());
}

void MP::start(const Ice::DoubleSeq &goals)
{
    start(goals, std::vector<double>());
}

void MP::start(const Ice::DoubleSeq& goals, Ice::Double timeDuration)
{
    cfg.durationSec = timeDuration;
    start(goals);
}

void MP::start(const Ice::DoubleSeq &goals, const Ice::DoubleSeq &starts, Ice::Double timeDuration)
{
    cfg.durationSec = timeDuration;
    start(goals, starts);
}

void MP::start(const Ice::DoubleSeq &goals, const Ice::DoubleSeq &starts)
{
    ARMARX_INFO << "--------------------------------------- start ---------------------------------------";
    canonicalValue = 1.0;
    /// parse user defined viapoints from configuration file
    for (const auto& vp: userDefinedViaPoints)
    {
        setViaPoint(vp.first, vp.second);
        //        ARMARX_INFO << "viapoint at " << vp.first << " is " << common::dVecToString(vp.second);
    }

    /// overwrite goals and starts if available
    setStartAndGoal(starts, goals);

    /// prepare execution
    LockGuardType guard{mpMutex};
    vmp->prepareExecution(targetPoseVec, currentState);
    auto vps = vmp->getViaPoints();
    ARMARX_INFO << "Your MP have: ";
    for (const auto& v: vps)
    {
        ARMARX_INFO << "viapoint at " << v.first << " is " << common::dVecToString(mplib::core::SystemState::convertStatesToArray(v.second, 0));
    }

    /// set flag
    running.store(true);
    resume();
    ARMARX_INFO << "--------------------------------------- started ---------------------------------------";
}

void MP::stop()
{
    running.store(false);
}

void MP::pause()
{
    paused.store(true);
}

void MP::resume()
{
    paused.store(false);
}

void MP::reset()
{
    if (running.load())
    {
        ARMARX_INFO << "cannot reset a running mp, please stop it first.";
    }
    else
    {
        userDefinedViaPoints.clear();
        targetPoseVec.clear();
        targetPoseVecInTraj.clear();
        currentState.clear();
        removeAllViaPoint();
        goalSetByUser.store(false);
        mpTrained.store(false);
    }
}

bool MP::isFinished()
{
    return !running.load();
}

void MP::learnFromCSV(const Ice::StringSeq& fileNames)
{
    ARMARX_IMPORTANT << "Train MP '" << cfg.name << "' from csv files";
    std::vector<std::string> fileList;
    if (fileNames.empty())
    {
        ARMARX_INFO << "using default trajectory list: " << cfg.fileList[0];
        fileList = cfg.fileList;
    }
    else
    {
        fileList = fileNames;
    }

    std::vector<mplib::core::SampledTrajectory> trajs;
    for (auto& file : fileList)
    {
        mplib::core::SampledTrajectory traj;
        traj.readFromCSVFile(file);
        trajs.push_back(traj);
    }

    if (cfg.regressionModel == "gpr")
    {
        vmp->setBaseFunctionApproximator(
                    std::make_unique<mplib::math::function_approximation::GaussianProcessRegression<mplib::math::kernel::SquaredExponentialCovarianceFunction>>
                    (0.01, 0.1, 0.000001)
                    );
        for (auto& traj : trajs)
        {
            traj = mplib::core::SampledTrajectory::downSample(traj, 100);
        }
    }

    vmp->learnFromTrajectories(trajs);

    currentState.clear();
    targetPoseVecInTraj.clear();

    ARMARX_INFO << "Your trajectory has " << VAROUT(trajs[0].dim());
    for (size_t i = 0; i < trajs[0].dim(); i++)
    {
        mplib::representation::MPState state(trajs[0].begin()->getPosition(i), trajs[0].begin()->getDeriv(i, 1));
        currentState.push_back(state);
        targetPoseVecInTraj.push_back(trajs[0].rbegin()->getPosition(i));
    }
    mpTrained.store(true);
    ARMARX_INFO << "MP trained";
}

void MP::setGoal(const Ice::DoubleSeq& goals)
{
    ARMARX_INFO << "Reset MP goal: " << common::dVecToString(goals);
    if (!goals.empty())
    {
        goalSetByUser.store(true);
        targetPoseVec = goals;
    }
    else if (!targetPoseVec.empty())
    {
        ARMARX_INFO << "user specified empty goal, try to use goals in your configuration file:\n" << common::dVecToString(targetPoseVec);
    }
    else if (!targetPoseVecInTraj.empty())
    {
        ARMARX_INFO << "user specified empty goal, try to use goals learned from trajectory:\n" << common::dVecToString(targetPoseVecInTraj);
        targetPoseVec = targetPoseVecInTraj;
    }
    else
    {
        ARMARX_WARNING << "user specified empty goal";
    }
}

void MP::setStart(const Ice::DoubleSeq& starts)
{
    ARMARX_INFO << "Reset MP start: " << common::dVecToString(starts);
    if (!starts.empty())
    {
        ARMARX_CHECK_EQUAL(starts.size(), currentState.size()) << ": size of starts and learned mp are not consistent";
        for (size_t i = 0; i < currentState.size(); i++)
        {
            currentState[i].pos = starts[i];
            currentState[i].vel = 0.0;
        }
    }
    else
    {
        ARMARX_IMPORTANT << "user doesn't define start point, fall back to start point learned from trajectory";
    }
}

void MP::setStartAndGoal(const Ice::DoubleSeq& starts, const Ice::DoubleSeq& goals)
{
    setStart(starts);
    setGoal(goals);
}

void MP::setViaPoint(Ice::Double canVal, const Ice::DoubleSeq& viapoint)
{
    if (mpTrained.load())
    {
        LockGuardType guard{mpMutex};
        if (canVal <= vmp->getUmin())
        {
            setGoal(viapoint);
        }
        else
        {
            vmp->setViaPoint(canVal, viapoint);
        }
    }
    else
    {
        ARMARX_ERROR << "mp is not trained yet, potential memory allocation error! Please train mp before setting any via points";
    }
}

void MP::removeAllViaPoint()
{
    LockGuardType guard{mpMutex};
    vmp->removeViaPoints();
}

double MP::getCanonicalValue()
{
    return canonicalValue;
}

/// serialze
void MP::setWeight(const std::vector<std::vector<double>>& weights)
{
    vmp->setWeights(weights);
}

DoubleSeqSeq MP::getWeight()
{
    return vmp->getWeights();
}

std::string MP::serialize(const std::string& mode)
{
    //    std::stringstream ss;
    //    if (mode == "string")
    //    {
    //        boost::archive::text_oarchive oa{ss};
    //        oa << vmp.get();
    //    }
    //    else if (mode == "xml")
    //    {
    //        boost::archive::xml_oarchive oa(ss);
    //        oa << vmp.get();
    //    }
    //    return ss.str();
    return "";
}

Ice::DoubleSeq MP::deserialize(const std::string& mpString, const std::string &mode)
{
    //    std::stringstream ss;
    //    ss.str(mpString);
    //    mplib::representation::vmp::PrincipalComponentVMP* dmpPointer;
    //    if (mode == "string")
    //    {
    //        boost::archive::text_iarchive ia{ss};
    //        ia >> dmpPointer;
    //    }
    //    else if (mode == "xml")
    //    {
    //        boost::archive::xml_iarchive ia(ss);
    //        ia >> dmpPointer;
    //    }
    //    else
    //    {
    //        ARMARX_ERROR << VAROUT(mode) << " unsupported! please use 'string' or 'xml' mode";
    //    }
    //    vmp.reset(dmpPointer);
    return vmp->getGoals();
}

} /// namespace armarx::control::common::mp



