/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ...
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/IK/DifferentialIK.h>

#include <RobotAPI/components/units/RobotUnit/RobotUnit.h>
#include <RobotAPI/components/units/RobotUnit/NJointControllers/NJointController.h>
#include <RobotAPI/components/units/RobotUnit/ControlTargets/ControlTarget1DoFActuator.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValue1DoFActuator.h>

#include <armarx/control/common/FTSensor.h>
#include <armarx/control/common/control_law/KeypointsAdmittanceController.h>
#include <armarx/control/njoint_mp_controller/task_space/ControllerInterface.h>

#include <mplib/core/SampledTrajectory.h>
#include <mplib/representation/vmp/PrincipalComponentVMP.h>
#include <mplib/factories/VMPFactory.h>
#include <mplib/factories/AbstractMPFactory.h>
#include <mplib/factories/MPFactoryCreator.h>


namespace armarx::control::njoint_mp_controller::task_space
{
    namespace common = armarx::control::common;
    namespace law = armarx::control::common::control_law;

    /**
    * @defgroup Library-KeypointMPController KeypointMPController
    * @ingroup Library-RobotUnit-NJointControllers
    * A description of the library KeypointMPController.
    *
    * @class KeypointMPController
    * @ingroup Library-KeypointMPController
    * @brief Brief description of class KeypointMPController.
    *
    * Detailed description of class KeypointMPController.
    */
    class KeypointMPController :
        public NJointControllerWithTripleBuffer<law::TaskspaceKeypointsAdmittanceController::Config>,
        public KeypointMPControllerInterface
    {
    public:
        using ConfigPtrT = KeypointMPControllerConfigPtr;
        using VMPPtr = std::shared_ptr<mplib::representation::vmp::PrincipalComponentVMP>;
        //        using VMPPtr = std::shared_ptr<mplib::representation::AbstractMovementPrimitive>;

        KeypointMPController(const RobotUnitPtr& robotUnit, const NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&);
        std::string getClassName(const Ice::Current&) const override;
        void rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override;

    protected:
        // NJointController interface
        void rtPreActivateController() override;
        void onInitNJointController() override;

    public:
        // KeypointMPControllerInterface interface
        void setKeypoints(const Eigen::VectorXf&, const Ice::Current &) override;
        void resetKeypoints(const int n_points, const float keypoint_stiffness, const bool is_rigid, const Eigen::VectorXf& ctrl_mask, const Eigen::VectorXf& init_keypoints_position, const Ice::Current &) override;

    public:
        // NJointTaskspaceAdmittanceControllerInterface interface
        std::string getKinematicChainName(const Ice::Current &) override;
        void setTCPPose(const Eigen::Matrix4f&, const Ice::Current&) override;
        //                void setNullspaceJointAngles(const Eigen::VectorXf&, const Ice::Current&) override;

        /// set control parameter
        void setControlParameters(const std::string&, const Eigen::VectorXf&, const Ice::Current&) override;
        void setForceTorqueBaseline(const Eigen::Vector3f&, const Eigen::Vector3f&, const Ice::Current&) override;

        /// ft sensor
        void toggleGravityCompensation(const bool toggle, const Ice::Current &) override;
        void setTCPMass(Ice::Float, const Ice::Current&) override;
        void setTCPCoMInFTFrame(const Eigen::Vector3f&, const Ice::Current&) override;
        void calibrateFTSensor(const Ice::Current&) override;

    public:
        // MPPoolInterface interface
        std::string getNames(const Ice::Current &) override;
        void start(const Ice::DoubleSeq &, const Ice::Current &) override;
        void startWithTime(const Ice::DoubleSeq &, Ice::Double, const Ice::Current &) override;
        void startAsTraj(const Ice::Current &) override;
        void startAsTrajWithTime(Ice::Double, const Ice::Current &) override;
        void stop(const Ice::Current &) override;
        void pause(const Ice::Current &) override;
        void resume(const Ice::Current &) override;
        void reset(const Ice::Current &) override;
        bool isFinished(const Ice::Current &) override;
        void learnFromCSV(const Ice::StringSeq &, const Ice::Current &) override;
        void setGoal(const Ice::DoubleSeq &, const Ice::Current &) override;
        void setStartAndGoal(const Ice::DoubleSeq &, const Ice::DoubleSeq &, const Ice::Current &) override;
        void setViaPoint(Ice::Double, const Ice::DoubleSeq &, const Ice::Current &) override;
        void removeAllViaPoint(const Ice::Current &) override;
        std::string serialize(const Ice::Current &) override;
        Ice::DoubleSeq deserialize(const std::string& mpAsString, const Ice::Current &) override;
        Ice::Double getCanVal(const Ice::Current &) override;

        void toggleMP(const bool enableMP, const Ice::Current &) override;
        bool getMPEnabled(const Ice::Current &) override;

    protected:
        void onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) override;

        void publishPose(StringVariantBaseMap& datafields, const std::string& name, const Eigen::Matrix4f& pose);
        void publishVec6f(StringVariantBaseMap& datafields, const std::string& name, const Eigen::Vector6f& vec);
        void publishVecXf(StringVariantBaseMap& datafields, const std::string& name, const Eigen::Vector6f& vec);

    private:
        /// devices
        std::vector<const SensorValue1DoFActuatorTorque*> torqueSensors;
        std::vector<const SensorValue1DoFActuatorVelocity*> velocitySensors;
        std::vector<const SensorValue1DoFActuatorPosition*> positionSensors;
        std::vector<ControlTarget1DoFActuatorTorque*> targets;

        /// set buffers
        struct debugRTInfo
        {
            StringFloatDictionary desired_torques;
        };
        TripleBuffer<debugRTInfo> debugRTBuffer;
        TripleBuffer<law::TaskspaceKeypointsAdmittanceController::Status> controlStatusBuffer;
        TripleBuffer<common::FTSensor::FTBufferData> ftSensorBuffer;

        struct rt2mpInfo
        {
            mplib::core::DVec2d currentState;
            double deltaT;
        };
        TripleBuffer<rt2mpInfo> rt2mpBuffer;

        /// variables
        std::string kinematicChainName;
        std::vector<std::string> jointNames;
        law::TaskspaceKeypointsAdmittanceController controller;
        common::FTSensor ftsensor;

        std::atomic_bool rtFirstRun = true;
        std::atomic_bool rtReady = false;

        /// vmp
        VMPPtr pointVMP;
        Eigen::VectorXf initialStateEigen;
        mplib::core::DVec2d initialState;
        mplib::core::DVec2d currentState;
        mplib::core::DVec targetPosition;
        mplib::core::DVec goal;
        int numPoints = 0;
        void runMP();
        double canVal = 1.0;
        double timeDuration = 0.0;
        std::atomic_bool mpRunning = false;
        std::atomic_bool mpFinished = false;

        std::atomic_bool mpEnabled {false};

        // NJointTaskspaceAdmittanceControllerInterface interface
    public:
        void reconfigureController(const std::string &, const Ice::Current &) override;
    };
}
